# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Massimiliano Iacono
         Sim Bamford

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Use kivy to create an app which can receive data dicts as imported by bimvee
importAe, and allow synchronised playback for each of the contained channels and datatypes. 
"""
# standard imports 
import numpy as np
import sys, os

os.environ['KIVY_NO_ARGS'] = 'T'

# Optional import of tkinter allows setting of app size wrt screen size
try:
    import tkinter as tk
    from kivy.config import Config

    root = tk.Tk()
    screen_width = root.winfo_screenwidth()
    screen_height = root.winfo_screenheight()
    Config.set('graphics', 'position', 'custom')
    Config.set('graphics', 'left', int(screen_width / 8))
    Config.set('graphics', 'top', int(screen_width / 8))
    Config.set('graphics', 'width', int(screen_width / 4 * 3))
    Config.set('graphics', 'height', int(screen_height / 4 * 3))
    # Config.set('graphics', 'fullscreen', 1)
except ModuleNotFoundError:
    pass

# kivy imports
from kivy.app import App
from kivy.uix.slider import Slider
from kivy.clock import Clock
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.textinput import TextInput
from kivy.uix.popup import Popup
from kivy.uix.checkbox import CheckBox
from kivy.properties import ObjectProperty
from kivy.properties import StringProperty, NumericProperty
from kivy.properties import DictProperty
from kivy.core.window import Window

# To get the graphics, set this as the current working directory
os.chdir(os.path.dirname(os.path.abspath(__file__)))
# local imports (from bimvee)
try:
    from visualiser import VisualiserDvs
    from visualiser import VisualiserFrame
    from visualiser import VisualiserPoint3
    from visualiser import VisualiserPose6q
    from visualiser import VisualiserBoundingBoxes
    from visualiser import VisualiserOpticFlow
    from visualiser import VisualiserImu
    from timestamps import getLastTimestamp
except ModuleNotFoundError:
    if __package__ is None or __package__ == '':
        sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    from bimvee.visualiser import VisualiserDvs
    from bimvee.visualiser import VisualiserFrame
    from bimvee.visualiser import VisualiserPoint3
    from bimvee.visualiser import VisualiserPose6q
    from bimvee.visualiser import VisualiserBoundingBoxes
    from bimvee.visualiser import VisualiserOpticFlow
    from bimvee.visualiser import VisualiserImu
    from bimvee.timestamps import getLastTimestamp
    from bimvee.visualiser import VisualiserSkeleton

from mustard.viewer import Viewer


class ErrorPopup(Popup):
    label_text = StringProperty(None)


class WarningPopup(Popup):
    label_text = StringProperty(None)


class TextInputPopup(Popup):
    label_text = StringProperty(None)


class LoadDialog(FloatLayout):
    load = ObjectProperty(None)
    cancel = ObjectProperty(None)
    load_path = StringProperty(None)


class DictEditor(GridLayout):
    dict = DictProperty(None)

    def on_dict(self, instance, value):
        # 2020_03_10 Sim: Why only import Spinner here? at the top of the file
        # it was causing a crash when starting in a thread - no idea why
        from kivy.uix.spinner import Spinner
        for n, topic in enumerate(sorted(value)):
            check_box = CheckBox()
            self.add_widget(check_box)
            self.add_widget(TextInput(text=str(n)))
            spinner = Spinner(values=['dvs', 'frame', 'pose6q', 'cam', 'imu', 'flowMap', 'skeleton'])
            if 'events' in topic:
                spinner.text = 'dvs'
                check_box.active = True
            elif 'image' in topic or 'depthmap' in topic:
                spinner.text = 'frame'
                check_box.active = True
            elif 'pose' in topic:
                spinner.text = 'pose6q'
                check_box.active = True
            elif 'flow' in topic:
                spinner.text = 'flowMap'
                check_box.active = True
            elif 'imu' in topic:
                spinner.text = 'imu'
                check_box.active = True
            elif 'skeleton' in topic:
                spinner.text = 'skeleton'
                check_box.active = True

            self.add_widget(spinner)
            self.add_widget(TextInput(text=topic))

    def get_dict(self):
        from collections import defaultdict
        out_dict = defaultdict(dict)
        for dict_row in np.array(self.children[::-1]).reshape((-1, self.cols))[1:]:
            if not dict_row[0].active:
                continue
            ch = dict_row[1].text
            type = dict_row[2].text
            data = dict_row[3].text
            out_dict[ch][type] = data
        return out_dict


class TemplateDialog(FloatLayout):
    template = DictProperty(None)
    cancel = ObjectProperty(None)
    load = ObjectProperty(None)


class DataController(GridLayout):
    ending_time = NumericProperty(.0)
    filePathOrName = StringProperty('')
    data_dict = ObjectProperty({})  # A bimvee-style container of channels

    def __init__(self, **kwargs):
        super(DataController, self).__init__(**kwargs)

    def update_children(self):
        for child in self.children:
            child.get_frame(self.time_value, self.time_window)

    def add_viewer_and_resize(self, data_dict, channel_name=''):
        visualisers = []
        settings = {}
        for data_type in data_dict.keys():
            settings[data_type] = {}
            if data_type == 'dvs':
                visualiser = VisualiserDvs(data_dict[data_type])
                settings[data_type] = visualiser.get_settings()
            elif data_type == 'frame':
                visualiser = VisualiserFrame(data_dict[data_type])
            elif data_type == 'pose6q':
                visualiser = VisualiserPose6q(data_dict[data_type])
                settings[data_type] = visualiser.get_settings()
                channel_name = channel_name + '\nred=x green=y, blue=z'
            elif data_type == 'point3':
                visualiser = VisualiserPoint3(data_dict[data_type])
                settings[data_type] = visualiser.get_settings()
            elif data_type == 'boundingBoxes':
                visualiser = VisualiserBoundingBoxes(data_dict[data_type])
                settings[data_type] = visualiser.get_settings()
            elif data_type == 'flowMap':
                visualiser = VisualiserOpticFlow(data_dict[data_type])
            elif data_type == 'imu':
                visualiser = VisualiserImu(data_dict[data_type])
                settings[data_type] = visualiser.get_settings()
                channel_name = channel_name + '\nred=x green=y, blue=z'
            elif data_type == 'skeleton':
                visualiser = VisualiserSkeleton(data_dict[data_type])
                settings[data_type] = visualiser.get_settings()
            else:
                print("Warning! {} is not a recognized data type. Ignoring.".format(data_type))
                continue
            visualisers.append(visualiser)
        if visualisers:
            new_viewer = Viewer()
            new_viewer.title = channel_name
            new_viewer.visualisers = visualisers
            new_viewer.settings = settings
            self.add_widget(new_viewer)

            self.cols = int(np.ceil(np.sqrt(len(self.children))))

    def add_viewer_for_each_channel_and_data_type(self, in_dict, seen_keys=[], recursionDepth=0):
        if isinstance(in_dict, list):
            print('    ' * recursionDepth + 'Received a list - looking through the list for containers...')
            for num, in_dict_element in enumerate(in_dict):
                seen_keys.append(num)
                self.add_viewer_for_each_channel_and_data_type(in_dict_element,
                                                               seen_keys=seen_keys,
                                                               recursionDepth=recursionDepth + 1)
        elif isinstance(in_dict, dict):
            print('    ' * recursionDepth + 'Received a dict - looking through its keys ...')
            for key_name in in_dict.keys():
                print('    ' * recursionDepth + 'Dict contains a key "' + key_name + '" ...')
                if isinstance(in_dict[key_name], dict):
                    seen_keys.append(key_name)
                    if 'ts' in in_dict[key_name]:
                        print('    ' * recursionDepth + 'Creating a new viewer, of type: ' + key_name)
                        self.add_viewer_and_resize(in_dict,
                                                   channel_name=seen_keys[-2])
                        break # We suppose that all timestamped data are at the same level
                    else:  # recurse through the sub-dict
                        self.add_viewer_for_each_channel_and_data_type(in_dict[key_name],
                                                                       seen_keys=seen_keys,
                                                                       recursionDepth=recursionDepth + 1)
                elif isinstance(in_dict[key_name], list):
                    self.add_viewer_for_each_channel_and_data_type(in_dict[key_name],
                                                                   seen_keys=seen_keys,
                                                                   recursionDepth=recursionDepth + 1)
                else:
                    print('    ' * recursionDepth + 'Ignoring that key ...')

    def on_data_dict(self, instance, value):
        while len(self.children) > 0:
            self.remove_widget(self.children[0])
            print('Removed an old viewer; num remaining viewers: ' + str(len(self.children)))
        if (self.data_dict is None) or not self.data_dict:
            # When using ntupleviz programmatically, pass an empty dict or None 
            # to allow the container to be passed again once updated
            return
        self.ending_time = float(getLastTimestamp(self.data_dict))  # timer is watching this
        self.add_viewer_for_each_channel_and_data_type(self.data_dict)

    def dismiss_popup(self):
        if hasattr(self, '_popup'):
            self._popup.dismiss()

    def show_template_dialog(self, topics):
        self.dismiss_popup()
        content = TemplateDialog(template=topics,
                                 cancel=self.dismiss_popup,
                                 load=self.load)
        self._popup = Popup(title="Define template", content=content,
                            size_hint=(0.9, 0.9))
        self._popup.open()

    def show_warning_popup(self, message):
        self.dismiss_popup()
        self._popup = WarningPopup(label_text=message,
                                   size_hint=(0.5, 0.5))
        self._popup.open()

    def show_load(self):
        self.dismiss_popup()
        content = LoadDialog(load=self.load,
                             cancel=self.dismiss_popup)
        self._popup = Popup(title="Load file", content=content,
                            size_hint=(0.9, 0.9))
        self._popup.open()

    def load(self, path, selection, template=None):
        self.dismiss_popup()
        try:
            from importAe import importAe
        except ModuleNotFoundError:
            from bimvee.importAe import importAe

        from os.path import join

        # If both path and selection are None than it will try to reload previously given path
        if path is not None or selection is not None:
            if selection:
                self.filePathOrName = join(path, selection[0])
            else:
                self.filePathOrName = path

        try:
            self.data_dict = importAe(filePathOrName=self.filePathOrName, template=template)
        except ValueError:
            try:
                from importRosbag import importRosbag
            except ModuleNotFoundError:
                from bimvee.importRosbag.importRosbag.importRosbag import importRosbag
            topics = importRosbag(filePathOrName=self.filePathOrName, listTopics=True)
            self.show_template_dialog(topics)
        except FileNotFoundError:
            self.show_warning_popup('Could not find any loadable dataset')
        self.update_children()


class TimeSlider(Slider):
    def __init__(self, **kwargs):
        super(TimeSlider, self).__init__(**kwargs)
        self.clock = None
        self.speed = 1

    def increase_slider(self, dt):
        self.value = min(self.value + dt / self.speed, self.max)
        if self.value >= self.max:
            if self.clock is not None:
                self.clock.cancel()

    def decrease_slider(self, dt):
        self.value = max(self.value - dt / self.speed, 0.0)
        if self.value <= 0.0:
            if self.clock is not None:
                self.clock.cancel()

    def play_pause(self):
        if self.clock is None:
            self.clock = Clock.schedule_interval(self.increase_slider, 0.001)
        else:
            if self.clock.is_triggered:
                self.clock.cancel()
            else:
                self.clock.cancel()
                self.clock = Clock.schedule_interval(self.increase_slider, 0.001)

    def pause(self):
        if self.clock is not None:
            self.clock.cancel()

    def play_forward(self):
        if self.clock is not None:
            self.clock.cancel()
        self.clock = Clock.schedule_interval(self.increase_slider, 0.001)

    def play_backward(self):
        if self.clock is not None:
            self.clock.cancel()
        self.clock = Clock.schedule_interval(self.decrease_slider, 0.001)

    def stop(self):
        if self.clock is not None:
            self.clock.cancel()
            self.set_norm_value(0)

    def reset(self):
        self.value = 0

    def step_forward(self):
        # self.increase_slider(self.time_window)
        self.increase_slider(0.016)

    def step_backward(self):
        # self.decrease_slider(self.time_window)
        self.decrease_slider(0.016)


class RootWidget(BoxLayout):
    def __init__(self, **kwargs):
        super(RootWidget, self).__init__(**kwargs)
        self._keyboard = Window.request_keyboard(self._keyboard_closed, self)
        self._keyboard.bind(on_key_down=self._on_keyboard_down, on_key_up=self._on_keyboard_up)

    def _on_keyboard_down(self, keyboard, keycode, text, modifiers):
        if keycode[1] == 'right':
            self.ids['time_slider'].step_forward()
        if keycode[1] == 'left':
            self.ids['time_slider'].step_backward()
        if keycode[1] == 'spacebar':
            self.ids['time_slider'].play_pause()
        for viewer in self.data_controller.children:
            viewer.transform_allowed = 'shift' in modifiers
            try:
                viewer.label = int(keycode[1][-1])
            except ValueError:
                continue
            # Return True to accept the key. Otherwise, it will be used by the system.
        return True

    def _on_keyboard_up(self, keyboard, keycode):
        for viewer in self.data_controller.children:
            viewer.transform_allowed = False
            # Return True to accept the key. Otherwise, it will be used by the system.
        return True

    def _keyboard_closed(self):
        self._keyboard.unbind(on_key_down=self._on_keyboard_down)
        del self._keyboard


class Mustard(App):
    def build(self):
        return RootWidget()

    def setData(self, newDataDict):
        self.root.data_controller.data_dict = {}
        self.root.data_controller.data_dict = newDataDict



# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Massimiliano Iacono
         Sim Bamford

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.

Collection of possible Viewers that can be spawned in the main app window. Each one will have
a visualiser which is responsible for the data management and retrieval.
"""

import numpy as np
import re
from kivy.graphics.texture import Texture
from kivy.uix.widget import Widget
from kivy.uix.slider import Slider
from kivy.uix.scatter import Scatter
from kivy.uix.checkbox import CheckBox
from kivy.uix.label import Label
from kivy.properties import BooleanProperty, StringProperty, ListProperty, DictProperty, NumericProperty
from kivy.metrics import dp
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout
from kivy.core.window import Window, Clock
from kivy.graphics.transformation import Matrix
from kivy.graphics.vertex_instructions import Rectangle

from bimvee.visualisers.visualiserBoundingBoxes import VisualiserBoundingBoxes
import os


class BoundingBox(Widget):
    mouse_over = BooleanProperty(False)

    def __init__(self, bb_color, x, y, width, height, **kwargs):
        super(BoundingBox, self).__init__(**kwargs)
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.bb_color = bb_color
        Window.bind(mouse_pos=self.on_mouse_pos)

    def on_mouse_pos(self, window, pos):
        win_coords = self.to_window(*self.pos)
        self.mouse_over = win_coords[0] < pos[0] < win_coords[0] + self.size[0] and\
                          win_coords[1] < pos[1] < win_coords[1] + self.size[1]


class LabeledBoundingBox(BoundingBox):
    obj_label = StringProperty('')

    def __init__(self, bb_color, x, y, width, height, label, **kwargs):
        super(LabeledBoundingBox, self).__init__(bb_color, x, y, width, height, **kwargs)
        try:
            self.obj_label = '{:d}'.format(int(label))
        except ValueError:
            self.obj_label = label


class ZoomableImage(Scatter):

    def __init__(self, **kwargs):
        super(ZoomableImage, self).__init__(**kwargs)
        Clock.schedule_once(self.reset_zoom)

    def reset_zoom(self, _):
        self.apply_transform(Matrix().scale(1, 1, 1),
                             anchor=self.center)

    def on_touch_down(self, touch):
        if not self.transform_allowed:
            return False
        if not (self.clickable_area[0] < touch.x < self.clickable_area[0] + self.clickable_area[2] and
                self.clickable_area[1] < touch.y < self.clickable_area[1] + self.clickable_area[3]):
            return False
        if touch.is_mouse_scrolling:
            factor = None
            if touch.button == 'scrolldown':
                if self.scale < self.scale_max:
                    factor = 1.1
            elif touch.button == 'scrollup':
                if self.scale > self.scale_min:
                    factor = 1 / 1.1
            if factor is not None:
                self.apply_transform(Matrix().scale(factor, factor, factor),
                                     anchor=touch.pos)
            return False
        else:
            return super(ZoomableImage, self).on_touch_down(touch)


class Viewer(BoxLayout):
    data = DictProperty(force_dispatch=True)
    visualisers = ListProperty([], allownone=True)
    flipHoriz = BooleanProperty(False)
    flipVert = BooleanProperty(False)
    labeling = BooleanProperty(False)
    transform_allowed = BooleanProperty(False)
    mouse_on_image = BooleanProperty(False)
    settings = DictProperty({}, allownone=True)
    settings_values = DictProperty({}, allownone=True)
    title = StringProperty('Title')
    colorfmt = StringProperty('luminance')
    orientation = 'vertical'
    mouse_position = ListProperty([0, 0])
    label = NumericProperty(0)
    len_gt = NumericProperty(0)

    def __init__(self, **kwargs):
        super(Viewer, self).__init__(**kwargs)
        self.settings_box = None
        from matplotlib.pyplot import get_cmap
        self.cm = get_cmap('tab20')
        self.current_time = 0
        self.current_time_window = 0
        Window.bind(mouse_pos=self.on_mouse_pos)
        self.clicked_mouse_pos = None
        self.last_added_box_idx = -1
        self.cropped_region = [0, 0, 0, 0]

    def window_to_image_coords(self, x, y, flip=True):
        scale = self.image.parent.scale
        scaled_image_width = self.image.norm_image_size[0] * scale
        scaled_image_height = self.image.norm_image_size[1] * scale
        w_ratio = scaled_image_width / self.image.texture.width
        h_ratio = scaled_image_height / self.image.texture.height
        window_img_x, window_img_y = self.image.to_window(self.image.center_x, self.image.center_y)
        image_x = ((x - (window_img_x - scaled_image_width / 2)) / w_ratio)
        image_y = ((y - (window_img_y - scaled_image_height / 2)) / h_ratio)

        if flip:
            if self.flipHoriz:
                image_x = self.image.texture.width - image_x
            if self.flipVert:
                image_y = self.image.texture.height - image_y
        return image_x, image_y

    def on_mouse_pos(self, window, pos):
        image_x, image_y = self.window_to_image_coords(pos[0], pos[1])
        self.mouse_on_image = True
        if self.cropped_region[0] <= image_x <= self.cropped_region[0] + self.cropped_region[2]:
            self.mouse_position[0] = int(image_x)
        elif image_x < self.cropped_region[0]:
            self.mouse_position[0] = int(self.cropped_region[0])
            self.mouse_on_image = False
        elif image_x > self.cropped_region[2]:
            self.mouse_position[0] = int(self.cropped_region[2])
            self.mouse_on_image = False
        if self.cropped_region[1] <= image_y <= self.cropped_region[1] + self.cropped_region[3]:
            self.mouse_position[1] = int(self.image.texture.height - image_y)
        elif image_y < self.cropped_region[1]:
            self.mouse_position[1] = int(self.cropped_region[3])
            self.mouse_on_image = False
        elif image_y > self.cropped_region[3]:
            self.mouse_position[1] = int(self.cropped_region[1])
            self.mouse_on_image = False

    def remove_last_bbox(self):
        for v in self.visualisers:
            if isinstance(v, VisualiserBoundingBoxes):
                data_dict = v.get_data()
                if self.last_added_box_idx != -1 and data_dict['orderAdded'][self.last_added_box_idx] != -1:
                    for d in data_dict:
                        try:
                            data_dict[d] = np.delete(data_dict[d], self.last_added_box_idx)
                        except IndexError:
                            return
                    try:
                        self.last_added_box_idx = np.argmax(data_dict['orderAdded'])
                    except ValueError:
                        self.last_added_box_idx = -1
                else:
                    return
                v.set_data(data_dict)
                self.len_gt = len(data_dict['ts'])
                self.get_frame(self.current_time, self.current_time_window)

    def save_bboxes(self, path, ending_time):
        if self.labeling:
            data_dict = None
            for v in self.visualisers:
                if isinstance(v, VisualiserBoundingBoxes):
                    data_dict = v.get_data()
                    viz = v
                    break
            if data_dict is None:
                return
            if self.settings_values[viz.data_type]['interpolate']:
                boxes = []
                for t in np.arange(0, ending_time, 0.01):  # TODO parametrize sample rate when saving interpolated
                    boxes_at_time = viz.get_frame(t, self.current_time_window, **self.settings_values[viz.data_type])
                    if boxes_at_time != [[0, 0, 0, 0]] and len(boxes_at_time):
                        for b in boxes_at_time:
                            boxes.append(np.concatenate(([t], b)))
            else:
                boxes = np.column_stack((data_dict['ts'], data_dict['minY'], data_dict['minX'], data_dict['maxY'],
                                         data_dict['maxX'], data_dict['label']))
            if not os.path.isdir(path):
                path = os.path.dirname(path)
            np.savetxt(os.path.join(path, 'ground_truth.csv'), boxes, fmt='%f')

    def on_touch_move(self, touch):
        if self.clicked_mouse_pos is not None:
            for v in self.visualisers:
                if isinstance(v, VisualiserBoundingBoxes):
                    data_dict = v.get_data()
                    viz = v
                    break

            data_dict['ts'][self.last_added_box_idx] = self.current_time
            data_dict['minY'][self.last_added_box_idx] = min(self.mouse_position[1], self.clicked_mouse_pos[1])
            data_dict['maxY'][self.last_added_box_idx] = max(self.mouse_position[1], self.clicked_mouse_pos[1])
            data_dict['minX'][self.last_added_box_idx] = min(self.mouse_position[0], self.clicked_mouse_pos[0])
            data_dict['maxX'][self.last_added_box_idx] = max(self.mouse_position[0], self.clicked_mouse_pos[0])
            viz.set_data(data_dict)
            self.get_frame(self.current_time, self.current_time_window)
        return False

    def on_touch_up(self, touch):
        self.clicked_mouse_pos = None
        for v in self.visualisers:
            if isinstance(v, VisualiserBoundingBoxes):
                data_dict = v.get_data()
                try:
                    if data_dict['minY'][-1] == data_dict['maxY'][-1] or data_dict['minX'][-1] == data_dict['maxX'][-1]:
                        self.remove_last_bbox()
                except IndexError:
                    pass
        return False

    def on_touch_down(self, touch):
        if super(Viewer, self).on_touch_down(touch):
            return True
        if not self.mouse_on_image:
            return False
        if self.transform_allowed:
            return False
        if self.labeling:
            data_dict = None
            viz = None
            for v in self.visualisers:
                if isinstance(v, VisualiserBoundingBoxes):
                    data_dict = v.get_data()
                    viz = v
                    break
            if data_dict is None:
                data_dict = {
                    'ts': np.array([self.current_time]),
                    'minY': np.array([self.mouse_position[1]]),
                    'minX': np.array([self.mouse_position[0]]),
                    'maxY': np.array([self.mouse_position[1]]),
                    'maxX': np.array([self.mouse_position[0]]),
                    'label': np.array([self.label]),
                    'orderAdded': np.array([0])
                }
                viz = VisualiserBoundingBoxes(data_dict)
                self.settings['boundingBoxes'] = viz.get_settings()
                self.visualisers.append(viz)
            else:
                data_dict['ts'] = np.append(data_dict['ts'], self.current_time)
                data_dict['minY'] = np.append(data_dict['minY'], self.mouse_position[1])
                data_dict['minX'] = np.append(data_dict['minX'], self.mouse_position[0])
                data_dict['maxY'] = np.append(data_dict['maxY'], self.mouse_position[1])
                data_dict['maxX'] = np.append(data_dict['maxX'], self.mouse_position[0])
                data_dict['label'] = np.append(data_dict['label'], self.label)
                try:
                    added_box = data_dict['orderAdded'].max() + 1
                except ValueError:
                    added_box = 0
                except KeyError:
                    data_dict['orderAdded'] = np.full(len(data_dict['ts']) - 1, -1)
                    added_box = 0

                data_dict['orderAdded'] = np.append(data_dict['orderAdded'], added_box)
            # Sorting wrt timestamps
            argsort = np.argsort(data_dict['ts'])
            for d in data_dict:
                if hasattr(data_dict[d], '__len__'):
                    data_dict[d] = data_dict[d][argsort]

            self.last_added_box_idx = np.argmax(data_dict['orderAdded'])
            self.len_gt = len(data_dict['ts'])
            viz.set_data(data_dict)
            self.get_frame(self.current_time, self.current_time_window)
            self.clicked_mouse_pos = self.mouse_position[0], self.mouse_position[1]
        return False

    def on_visualisers(self, instance, value):
        self.init_visualisers()

    def init_visualisers(self):
        if self.visualisers is not None and self.visualisers:
            for v in self.visualisers:
                if v.data_type in ['dvs', 'frame', 'pose6q', 'point3', 'flowMap', 'imu']:
                    self.colorfmt = v.get_colorfmt()
                    self.data_shape = v.get_dims()
                    buf_shape = (dp(self.data_shape[0]), dp(self.data_shape[1]))
                    self.image.texture = Texture.create(size=buf_shape, colorfmt=self.colorfmt)
                if v.data_type == 'boundingBoxes':
                    self.len_gt = len(v.get_data()['ts'])

    def on_settings(self, instance, settings_dict):
        if self.settings_box is not None:
            self.settings_box.clear_widgets()
        self.settings_box = BoxLayout(size_hint=(1, 0.2), spacing=5)
        self.add_widget(self.settings_box)
        self.update_settings(self.settings_box, settings_dict, self.settings_values)

    def on_settings_change(self, instance, value):
        self.settings_values[instance.parent.id][instance.id] = value
        self.get_frame(self.current_time, self.current_time_window)

    def update_settings(self, parent_widget, settings_dict, settings_values):
        for key in settings_dict:
            if 'type' not in settings_dict[key]:
                if settings_dict[key]:
                    box = BoxLayout(orientation='vertical', spacing=5)
                    splitted = re.sub('([A-Z][a-z]+)', r' \1', re.sub('([A-Z]+)', r' \1', key)).title()
                    box.add_widget(Label(text='Settings for {}'.format(splitted), size_hint=(1, 0.1)))
                    settings_grid = GridLayout(cols=2)
                    settings_grid.id = key
                    box.add_widget(settings_grid)
                    parent_widget.add_widget(box)
                    settings_values[key] = {}
                    self.update_settings(settings_grid, settings_dict[key], settings_values[key])
            elif settings_dict[key]['type'] == 'boolean':
                parent_widget.add_widget(Label(text=key))
                check_box = CheckBox(active=settings_dict[key]['default'])
                check_box.id = key
                parent_widget.add_widget(check_box)
                settings_values[key] = check_box.active
                check_box.bind(active=self.on_settings_change)
            elif settings_dict[key]['type'] == 'range':
                parent_widget.add_widget(Label(text=key))
                slider = Slider(value=settings_dict[key]['default'],
                                min=settings_dict[key]['min'],
                                max=settings_dict[key]['max'],
                                step=settings_dict[key]['step'])
                slider.id = key
                parent_widget.add_widget(slider)
                settings_values[key] = slider.value
                slider.bind(value=self.on_settings_change)
            elif settings_dict[key]['type'] == 'value_list':
                parent_widget.add_widget(Label(text=key))
                from kivy.uix.spinner import Spinner
                spinner = Spinner(text=settings_dict[key]['default'],
                                  values=settings_dict[key]['values'])
                spinner.id = key
                parent_widget.add_widget(spinner)
                settings_values[key] = spinner.text
                spinner.bind(text=self.on_settings_change)

    def on_data(self, instance, value):
        self.image.clear_widgets()
        for data_type in self.data.keys():
            if data_type in ['dvs', 'frame', 'pose6q', 'point3', 'flowMap', 'imu']:
                self.update_image(self.data[data_type])
            elif data_type in ['boundingBoxes']:
                self.update_b_boxes(self.data[data_type])
            elif data_type in ['skeleton']:
                self.update_skeleton(self.data[data_type])

    def update_image(self, data):
        if self.image.texture is not None:
            size_required = self.data_shape[0] * self.data_shape[1] * (1 + (self.colorfmt == 'rgb') * 2)
            if not isinstance(data, np.ndarray):
                data = np.zeros((self.data_shape[0], self.data_shape[1], 3), dtype=np.uint8)
            if data.size >= size_required:
                try:
                    if self.flipHoriz:
                        data = np.flip(data, axis=1)
                    if not self.flipVert:  # Not, because by default, y should increase downwards, following https://arxiv.org/pdf/1610.08336.pdf
                        data = np.flip(data, axis=0)
                except AttributeError:
                    pass  # It's not a class that allows flipping
                self.image.texture.blit_buffer(data.tostring(), bufferfmt="ubyte", colorfmt=self.colorfmt)

    def update_b_boxes(self, b_boxes):
        if b_boxes is None:
            return

        bb_copy = b_boxes.copy()
        texture_width = self.image.texture.width
        texture_height = self.image.texture.height
        x_img, y_img, image_width, image_height = self.get_image_bounding_box()

        w_ratio = image_width / texture_width
        h_ratio = image_height / texture_height
        for n, b in enumerate(bb_copy):
            for i in range(4):
                b[i] = dp(b[i])
            if self.flipHoriz:
                min_x = texture_width - b[3]
                max_x = texture_width - b[1]
                b[1] = min_x
                b[3] = max_x
            if self.flipVert:
                min_y = texture_height - b[2]
                max_y = texture_height - b[0]
                b[0] = min_y
                b[2] = max_y

            if not (self.cropped_region[0] < (b[3] + b[1]) / 2 < self.cropped_region[0] + self.cropped_region[2] and
                    self.cropped_region[1] < texture_height - ((b[2] + b[0]) / 2) < self.cropped_region[1] + self.cropped_region[3]):
                continue

            width = w_ratio * float(b[3] - b[1])
            height = h_ratio * float(b[2] - b[0])
            if width == 0 and height == 0:
                break

            x = x_img + w_ratio * float(b[1])
            y = y_img + h_ratio * (texture_height - float(b[2]))

            try:
                bb_color = self.cm.colors[b[4] % len(self.cm.colors)] + (1,)
                label = b[4]
                if label == 0:  # Label = 0 is considered as unlabeled
                    raise IndexError
                box_item = LabeledBoundingBox(bb_color=bb_color,
                                              x=x, y=y,
                                              width=width, height=height,
                                              label=label)
            except IndexError:
                box_item = BoundingBox(bb_color=self.cm.colors[0],
                                       x=x, y=y,
                                       width=width, height=height)

            box_item.id = 'box_{}'.format(n),

            self.image.add_widget(box_item)

    def update_skeleton(self, skeleton):
        if skeleton is None:
            return

        texture_width = self.image.texture.width
        texture_height = self.image.texture.height
        x_img, y_img, image_width, image_height = self.get_image_bounding_box()

        w_ratio = image_width / texture_width
        h_ratio = image_height / texture_height
        for i, joint in enumerate(skeleton):
            x = skeleton[joint][0]
            y = texture_height - skeleton[joint][1]

            if self.flipHoriz:
                x = texture_width - x
            if self.flipVert:
                y = texture_height - y

            if not (self.cropped_region[0] < x < self.cropped_region[0] + self.cropped_region[2] and
                    self.cropped_region[1] < y < self.cropped_region[1] + self.cropped_region[3]):
                continue

            x = int(x_img + w_ratio * x)
            y = int(y_img + h_ratio * y)

            if self.settings_values['skeleton']['show_labels']:
                box_item = LabeledBoundingBox(bb_color=self.cm.colors[i % len(self.cm.colors)] + (1,),
                                              x=x, y=y,
                                              width=2, height=2,
                                              label=joint)
            else:
                box_item = BoundingBox(bb_color=self.cm.colors[i % len(self.cm.colors)] + (1,),
                                       x=x, y=y,
                                       width=2, height=2)
            box_item.id = 'box_{}'.format(i),

            self.image.add_widget(box_item)

    def get_image_bounding_box(self):
        image_width = self.image.norm_image_size[0]
        image_height = self.image.norm_image_size[1]
        x_img = self.image.center_x - image_width / 2
        y_img = self.image.center_y - image_height / 2
        return x_img, y_img, image_width, image_height

    def get_frame(self, time_value, time_window):
        data_dict = {}
        self.current_time = time_value
        self.current_time_window = time_window
        for v in self.visualisers:
            data_dict[v.data_type] = {}
            data_dict[v.data_type] = v.get_frame(time_value, time_window, **self.settings_values[v.data_type])
            self.colorfmt = v.get_colorfmt()
        self.data.update(data_dict)

    def crop_image(self, x, y, width, height):
        _, _, image_width, image_height = self.get_image_bounding_box()

        crop_bl_x, crop_bl_y = self.window_to_image_coords(x, y, flip=False)
        crop_tr_x, crop_tr_y = self.window_to_image_coords(x + width, y + height, flip=False)

        texture_width = self.image.texture.width
        texture_height = self.image.texture.height
        crop_bl_x = max(0, crop_bl_x)
        crop_bl_x = min(texture_width, crop_bl_x)
        crop_bl_y = max(0, crop_bl_y)
        crop_bl_y = min(texture_height, crop_bl_y)
        crop_width = min(crop_tr_x - crop_bl_x, texture_width - crop_bl_x)
        crop_height = min(crop_tr_y - crop_bl_y, texture_height - crop_bl_y)

        w_ratio = image_width / texture_width
        h_ratio = image_height / texture_height

        subtexture = self.image.texture.get_region(crop_bl_x, crop_bl_y, crop_width, crop_height)

        self.cropped_region = [crop_bl_x, crop_bl_y, crop_width, crop_height]

        with self.image.canvas:
            self.image.canvas.clear()
            Rectangle(texture=subtexture,
                      pos=(crop_bl_x * w_ratio + self.image.center_x - image_width / 2,
                           crop_bl_y * h_ratio + self.image.center_y - image_height / 2),
                      size=(crop_width * w_ratio, crop_height * h_ratio))
        self.on_data(None, None)



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='mustard-gui',
    packages=find_packages(),
    version='1.0.6',
    license='gpl',
    description='MUlti STream Agnostic Representation Dataplayer',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Event-driven Perception for Robotics group at Istituto Italiano di Tecnologia: Massimiliano Iacono',
    author_email='massimiliano.iacono@iit.it',
    url='https://github.com/event-driven-robotics/mustard',
    keywords=['event', 'event camera', 'event-based', 'event-driven', 'spike', 'dvs', 'dynamic vision sensor',
              'neuromorphic', 'aer', 'address-event representation' 'spiking neural network', 'davis', 'atis', 'celex'],
    install_requires=['kivy>=2.0.0',
                      'matplotlib',
                      'numpy',
                      'tqdm',
                      'bimvee>=1.0.18',
                      'scipy'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
    include_package_data=True,
    scripts=['mustard/mustard']
)
